{% extends "layout.html" %}
{% block body %}
<h1>Urwid{% if 'dev' in release %} development version{% endif %}</h1>
<div style="float: left; width: 500px;">
<p>Console user interface library for Python</p>

<ul>
{% if 'dev' in release %}<li>Stable release available at <a href="https://urwid.org/">https://urwid.org</a></li>
{% else %}
<li><a href="https://pypi.org/packages/source/u/urwid/urwid-{{ release }}.tar.gz"
>Download: urwid-{{ release }}.tar.gz</a></li>
{% endif %}
<li><a href="changelog.html">Changelog</a>
</ul>

<div class="hilight"><pre>git clone https://github.com/urwid/urwid.git</pre></div>

<ul>
<li><a href="https://github.com/urwid/urwid">Github page</a></li>
<li><a href="https://github.com/urwid/urwid/issues">Issues</a></li>
<li><a href="https://gitter.im/urwid/community">IRC: <code>#urwid/community on gitter.im</code></a>
</ul>

<p>Wiki:</p>

<ul>
<li><a href="https://github.com/urwid/urwid/wiki/Installation-instructions">Installation instructions</a></li>
<li><a href="https://github.com/urwid/urwid/wiki/FAQ">FAQ</a></li>
<li><a href="https://github.com/urwid/urwid/wiki/Application-list">Applications built with Urwid</a></li>
<li><a href="https://github.com/urwid/urwid/wiki/How-you-can-help">How you can help</a></li>
</ul>

<div class="section" id="requirements">
<h3>Requirements</h3>
<ul>
<li>Python 3.7+ or PyPy</li>
<li>Linux, OSX, Cygwin or other unix-like OS</li>
<li>python-gi for GlibEventLoop (optional)</li>
<li>Twisted for TwistedEventLoop (optional)</li>
<li>Tornado for TornadoEventLoop (optional)</li>
<li>Trio for TrioEventLoop (optional)</li>
<li>ZMQ for ZMQEventLoop (optional)</li>
<li>Apache for web_display module (optional)</li>
<li>ncurses for curses_display module (optional)</li>
</ul>
</div>

<div class="section" id="similar">
<h3>Similar projects</h3>
<ul>
<li><a href="https://www.npcole.com/npyscreen/">npyscreen</a></li>
<li><a href="https://github.com/thomasballinger/curtsies">curtsies</a></li>
<li><a href="https://github.com/jonathanslenders/python-prompt-toolkit">Python Prompt Toolkit</a></li>
</ul>
</div>
</div>

<div style="padding-left: 560px;">
<div class="section" id="big-buttons">
<p><a class="reference external image-reference" href="examples/index.html"><img alt="_images/graph2.png" src="_images/graph2.png" width="200" style="padding-bottom: 5px" /></br>
Examples</a></p>
<p><a class="reference external image-reference" href="tutorial/index.html"><img alt="_images/highcolors1.png" src="_images/highcolors1.png" width="200" style="padding-bottom: 5px" /></br>
Tutorial</a></p>
<p><a class="reference external image-reference" href="manual/index.html"><img alt="_images/introduction.png" src="_images/introduction.png" width="200" style="padding-bottom: 5px" /></br>
Manual</a></p>
<p><a class="reference external image-reference" href="reference/index.html"><img alt="_images/urwid_widgets_1.png" src="_images/urwid_widgets_1.png" width="200" style="padding-bottom: 5px" /></br>
Reference</a></p>
<p><a class="reference internal" href="genindex.html">Index</a></p>
<p><a class="reference internal" href="search.html">Search Page</a></p>
</div>
</div>
<div style="clear:both">&nbsp;</div>
{% endblock %}
